@@ -1,5 +1,5 @@
 /* gifread.c - Functions to read GIFs.
-   Copyright (C) 1997-2013 Eddie Kohler, kohler@cs.ucla.edu
+   Copyright (C) 1997-2014 Eddie Kohler, ekohler@gmail.com
    This file is part of the LCDF GIF library.
 
    The LCDF GIF library is free software. It is distributed under the GNU
@@ -37,8 +37,9 @@ typedef struct {
 
   unsigned decodepos;
 
+  Gif_Image* gfi;
   Gif_ReadErrorHandler handler;
-  void *handler_thunk;
+  int errors[2];
 
 } Gif_Context;
 
@@ -52,12 +53,14 @@ typedef struct Gif_Reader {
   int is_record;
   int is_eoi;
   uint8_t (*byte_getter)(struct Gif_Reader *);
-  void (*block_getter)(uint8_t *, uint32_t, struct Gif_Reader *);
+  uint32_t (*block_getter)(uint8_t*, uint32_t, struct Gif_Reader*);
   uint32_t (*offseter)(struct Gif_Reader *);
   int (*eofer)(struct Gif_Reader *);
 
 } Gif_Reader;
 
+static Gif_ReadErrorHandler default_error_handler = 0;
+
 
 #define gifgetc(grr)	((char)(*grr->byte_getter)(grr))
 #define gifgetbyte(grr) ((*grr->byte_getter)(grr))
@@ -77,22 +80,23 @@ gifgetunsigned(Gif_Reader *grr)
 static uint8_t
 file_byte_getter(Gif_Reader *grr)
 {
-  int i = getc(grr->f);
-  return i == EOF ? 0 : (uint8_t)i;
+    int i = getc(grr->f);
+    return i == EOF ? 0 : (uint8_t)i;
 }
 
-static void
+static uint32_t
 file_block_getter(uint8_t *p, uint32_t s, Gif_Reader *grr)
 {
-  size_t nread = fread(p, 1, s, grr->f);
-  if (nread < s)
-    memset(p + nread, 0, s - nread);
+    size_t nread = fread(p, 1, s, grr->f);
+    if (nread < s)
+        memset(p + nread, 0, s - nread);
+    return nread;
 }
 
 static uint32_t
 file_offseter(Gif_Reader *grr)
 {
-  return ftell(grr->f);
+    return ftell(grr->f);
 }
 
 static int
@@ -114,15 +118,16 @@ record_byte_getter(Gif_Reader *grr)
   return grr->w ? (grr->w--, *grr->v++) : 0;
 }
 
-static void
+static uint32_t
 record_block_getter(uint8_t *p, uint32_t s, Gif_Reader *grr)
 {
-  uint32_t ncopy = s;
-  if (ncopy > grr->w) ncopy = grr->w;
-  memcpy(p, grr->v, ncopy);
-  grr->w -= ncopy, grr->v += ncopy;
-  if (ncopy < s)
-    memset(p + ncopy, 0, s - ncopy);
+    uint32_t ncopy = (s <= grr->w ? s : grr->w);
+    memcpy(p, grr->v, ncopy);
+    grr->w -= ncopy;
+    grr->v += ncopy;
+    if (ncopy < s)
+        memset(p + ncopy, 0, s - ncopy);
+    return ncopy;
 }
 
 static uint32_t
@@ -155,9 +160,10 @@ make_data_reader(Gif_Reader *grr, const uint8_t *data, uint32_t length)
 static void
 gif_read_error(Gif_Context *gfc, int is_error, const char *text)
 {
-  gfc->stream->errors++;
-  if (gfc->handler)
-      gfc->handler(is_error, text, gfc->stream->nimages, gfc->handler_thunk);
+    Gif_ReadErrorHandler handler = gfc->handler ? gfc->handler : default_error_handler;
+    gfc->errors[is_error > 0] += 1;
+    if (handler)
+        handler(gfc->stream, gfc->gfi, is_error, text);
 }
 
 
@@ -167,28 +173,20 @@ one_code(Gif_Context *gfc, Gif_Code code)
   uint8_t *suffixes = gfc->suffix;
   Gif_Code *prefixes = gfc->prefix;
   uint8_t *ptr;
-  int lastsuffix;
-  uint16_t codelength = gfc->length[code];
+  int lastsuffix = 0;
+  int codelength = gfc->length[code];
 
   gfc->decodepos += codelength;
   ptr = gfc->image + gfc->decodepos;
-  if (ptr > gfc->maximage || !codelength) {
-    gif_read_error(gfc, 1, (!codelength ? "bad code" : "too much image data"));
-    /* 5/26/98 It's not good enough simply to count an error, because in the
-       read_image_data function, if code == next_code, we will store a byte in
-       gfc->image[gfc->decodepos-1]. Thus, fix decodepos so it's w/in the
-       image. */
-    gfc->decodepos = gfc->maximage - gfc->image;
-    return 0;
+  while (codelength > 0) {
+      lastsuffix = suffixes[code];
+      code = prefixes[code];
+      --ptr;
+      if (ptr < gfc->maximage)
+          *ptr = lastsuffix;
+      --codelength;
   }
 
-  /* codelength will always be greater than 0. */
-  do {
-    lastsuffix = suffixes[code];
-    *--ptr = lastsuffix;
-    code = prefixes[code];
-  } while (--codelength > 0);
-
   /* return the first pixel in the code, which, since we walked backwards
      through the code, was the last suffix we processed. */
   return lastsuffix;
@@ -253,10 +251,10 @@ read_image_data(Gif_Context *gfc, Gif_Reader *grr)
   min_code_size = gifgetbyte(grr);
   GIF_DEBUG(("\n\nmin_code_size(%d)", min_code_size));
   if (min_code_size >= GIF_MAX_CODE_BITS) {
-    gif_read_error(gfc, 1, "min_code_size too big");
+    gif_read_error(gfc, 1, "image corrupted, min_code_size too big");
     min_code_size = GIF_MAX_CODE_BITS - 1;
   } else if (min_code_size < 2) {
-    gif_read_error(gfc, 1, "min_code_size too small");
+    gif_read_error(gfc, 1, "image corrupted, min_code_size too small");
     min_code_size = 2;
   }
   clear_code = 1 << min_code_size;
@@ -326,7 +324,10 @@ read_image_data(Gif_Context *gfc, Gif_Reader *grr)
 	 Bug fix, 4/12/2010: It is not an error if next_code == clear_code.
 	 This happens at the end of a large GIF: see the next comment ("If no
 	 meaningful next code should be defined...."). */
-      gif_read_error(gfc, 1, "unexpected code");
+      if (gfc->errors[1] < 20)
+          gif_read_error(gfc, 1, "image corrupted, code out of range");
+      else if (gfc->errors[1] == 20)
+          gif_read_error(gfc, 1, "(not reporting more errors)");
       code = 0;
     }
 
@@ -346,7 +347,8 @@ read_image_data(Gif_Context *gfc, Gif_Reader *grr)
 
     /* Special processing if code == next_code: we didn't know code's final
        suffix when we called one_code, but we do now. */
-    if (code == next_code)
+    /* 7.Mar.2014 -- Avoid error if image has zero width/height. */
+    if (code == next_code && gfc->image + gfc->decodepos <= gfc->maximage)
       gfc->image[gfc->decodepos - 1] = gfc->suffix[next_code];
 
     /* Increment next_code except for the 'clear_code' special case (that's
@@ -373,12 +375,19 @@ read_image_data(Gif_Context *gfc, Gif_Reader *grr)
   }
 
   /* zero-length block reached. */
- zero_length_block:
-
-  if (gfc->image + gfc->decodepos < gfc->maximage)
-    gif_read_error(gfc, 1, "not enough image data for image size");
-  else if (gfc->image + gfc->decodepos > gfc->maximage)
-    gif_read_error(gfc, 1, "too much image data for image size");
+ zero_length_block: {
+      long delta = (long) (gfc->maximage - gfc->image) - (long) gfc->decodepos;
+      char buf[BUFSIZ];
+      if (delta > 0) {
+          sprintf(buf, "missing %ld pixels of image data", delta);
+          gif_read_error(gfc, 1, buf);
+      } else if (delta < -1) {
+          /* One pixel of superfluous data is OK; that could be the
+             code == next_code case. */
+          sprintf(buf, "%ld superfluous pixels of image data", -delta);
+          gif_read_error(gfc, 0, buf);
+      }
+  }
 }
 
 
@@ -450,7 +459,7 @@ read_compressed_image(Gif_Image *gfi, Gif_Reader *grr, int read_flags)
       gfi->free_compressed = 0;
     } else {
       gfi->compressed = Gif_NewArray(uint8_t, gfi->compressed_len);
-      gfi->free_compressed = Gif_DeleteArrayFunc;
+      gfi->free_compressed = Gif_Free;
       if (!gfi->compressed) return 0;
       memcpy(gfi->compressed, first, gfi->compressed_len);
     }
@@ -490,7 +499,7 @@ read_compressed_image(Gif_Image *gfi, Gif_Reader *grr, int read_flags)
 
     gfi->compressed = comp;
     gfi->compressed_len = comp_len;
-    gfi->free_compressed = Gif_DeleteArrayFunc;
+    gfi->free_compressed = Gif_Free;
   }
 
   return 1;
@@ -500,21 +509,22 @@ read_compressed_image(Gif_Image *gfi, Gif_Reader *grr, int read_flags)
 static int
 uncompress_image(Gif_Context *gfc, Gif_Image *gfi, Gif_Reader *grr)
 {
-  if (!Gif_CreateUncompressedImage(gfi)) return 0;
-  gfc->width = gfi->width;
-  gfc->height = gfi->height;
-  gfc->image = gfi->image_data;
-  gfc->maximage = gfi->image_data + gfi->width * gfi->height;
-  read_image_data(gfc, grr);
-  return 1;
+    if (!Gif_CreateUncompressedImage(gfi, gfi->interlace))
+        return 0;
+    gfc->width = gfi->width;
+    gfc->height = gfi->height;
+    gfc->image = gfi->image_data;
+    gfc->maximage = gfi->image_data + (unsigned) gfi->width * (unsigned) gfi->height;
+    read_image_data(gfc, grr);
+    return 1;
 }
 
 
 int
-Gif_FullUncompressImage(Gif_Image *gfi, Gif_ReadErrorHandler h, void *hthunk)
+Gif_FullUncompressImage(Gif_Stream* gfs, Gif_Image* gfi,
+                        Gif_ReadErrorHandler h)
 {
   Gif_Context gfc;
-  Gif_Stream fake_gfs;
   Gif_Reader grr;
   int ok = 0;
 
@@ -527,13 +537,13 @@ Gif_FullUncompressImage(Gif_Image *gfi, Gif_ReadErrorHandler h, void *hthunk)
        this shouldn't happen */
     return 0;
 
-  fake_gfs.errors = 0;
-  gfc.stream = &fake_gfs;
+  gfc.stream = gfs;
+  gfc.gfi = gfi;
   gfc.prefix = Gif_NewArray(Gif_Code, GIF_MAX_CODE);
   gfc.suffix = Gif_NewArray(uint8_t, GIF_MAX_CODE);
   gfc.length = Gif_NewArray(uint16_t, GIF_MAX_CODE);
   gfc.handler = h;
-  gfc.handler_thunk = hthunk;
+  gfc.errors[0] = gfc.errors[1] = 0;
 
   if (gfi && gfc.prefix && gfc.suffix && gfc.length && gfi->compressed) {
     make_data_reader(&grr, gfi->compressed, gfi->compressed_len);
@@ -543,7 +553,9 @@ Gif_FullUncompressImage(Gif_Image *gfi, Gif_ReadErrorHandler h, void *hthunk)
   Gif_DeleteArray(gfc.prefix);
   Gif_DeleteArray(gfc.suffix);
   Gif_DeleteArray(gfc.length);
-  return ok && !fake_gfs.errors;
+  if (gfc.errors[0] || gfc.errors[1])
+      gif_read_error(&gfc, -1, 0);
+  return ok && !gfc.errors[1];
 }
 
 
@@ -557,9 +569,28 @@ read_image(Gif_Reader *grr, Gif_Context *gfc, Gif_Image *gfi, int read_flags)
   gfi->top = gifgetunsigned(grr);
   gfi->width = gifgetunsigned(grr);
   gfi->height = gifgetunsigned(grr);
-  packed = gifgetbyte(grr);
+  /* Mainline GIF processors (Firefox, etc.) process missing width (height)
+     as screen_width (screen_height). */
+  if (gfi->width == 0)
+      gfi->width = gfc->stream->screen_width;
+  if (gfi->height == 0)
+      gfi->height = gfc->stream->screen_height;
+  /* If still zero, error. */
+  if (gfi->width == 0 || gfi->height == 0) {
+      gif_read_error(gfc, 1, "image has zero width and/or height");
+      Gif_MakeImageEmpty(gfi);
+      read_flags = 0;
+  }
+  /* If position out of range, error. */
+  if ((unsigned) gfi->left + (unsigned) gfi->width > 0xFFFF
+      || (unsigned) gfi->top + (unsigned) gfi->height > 0xFFFF) {
+      gif_read_error(gfc, 1, "image position and/or dimensions out of range");
+      Gif_MakeImageEmpty(gfi);
+      read_flags = 0;
+  }
   GIF_DEBUG(("<%ux%u>", gfi->width, gfi->height));
 
+  packed = gifgetbyte(grr);
   if (packed & 0x80) { /* have a local color table */
     int ncol = 1 << ((packed & 0x07) + 1);
     gfi->local = read_color_table(ncol, grr);
@@ -618,13 +649,13 @@ read_graphic_control_extension(Gif_Context *gfc, Gif_Image *gfi,
   }
 
   if (len > 0) {
-    gif_read_error(gfc, 1, "odd graphic extension format");
+    gif_read_error(gfc, 1, "bad graphic extension");
     gifgetblock(crap, len, grr);
   }
 
   len = gifgetbyte(grr);
   while (len > 0) {
-    gif_read_error(gfc, 1, "odd graphic extension format");
+    gif_read_error(gfc, 1, "bad graphic extension");
     gifgetblock(crap, len, grr);
     len = gifgetbyte(grr);
   }
@@ -657,41 +688,44 @@ suck_data(char *data, int *store_len, Gif_Reader *grr)
 
 
 static int
-read_unknown_extension(Gif_Stream *gfs, int kind, char *app_name, int position,
-		       Gif_Reader *grr)
+read_unknown_extension(Gif_Stream *gfs, int kind, char* appname, int applength,
+                       int position, Gif_Reader *grr)
 {
-  uint8_t block_len = gifgetbyte(grr);
-  uint8_t *data = 0;
-  uint8_t data_len = 0;
-  Gif_Extension *gfex = 0;
+    uint8_t block_len = gifgetbyte(grr);
+    uint8_t* data = 0;
+    int data_len = 0;
+    Gif_Extension *gfex = 0;
 
-  while (block_len > 0) {
-    if (data) Gif_ReArray(data, uint8_t, data_len + block_len + 1);
-    else data = Gif_NewArray(uint8_t, block_len + 1);
-    if (!data) goto done;
-    gifgetblock(data + data_len, block_len, grr);
-    data_len += block_len;
-    block_len = gifgetbyte(grr);
-  }
+    while (block_len > 0) {
+        Gif_ReArray(data, uint8_t, data_len + block_len + 2);
+        if (!data)
+            goto done;
+        data[data_len] = block_len;
+        gifgetblock(data + data_len + 1, block_len, grr);
+        data_len += block_len + 1;
+        block_len = gifgetbyte(grr);
+    }
 
-  if (data)
-    gfex = Gif_NewExtension(kind, app_name);
-  if (gfex) {
-    gfex->data = data;
-    gfex->free_data = Gif_DeleteArrayFunc;
-    gfex->length = data_len;
-    data[data_len] = 0;
-    Gif_AddExtension(gfs, gfex, position);
-  }
+    if (data)
+        gfex = Gif_NewExtension(kind, appname, applength);
+    if (gfex) {
+        gfex->data = data;
+        gfex->free_data = Gif_Free;
+        gfex->length = data_len;
+        gfex->packetized = 1;
+        data[data_len] = 0;
+        Gif_AddExtension(gfs, gfex, position);
+    }
 
  done:
-  if (!gfex) Gif_DeleteArray(data);
-  while (block_len > 0) {
-    uint8_t buffer[GIF_MAX_BLOCK];
-    gifgetblock(buffer, block_len, grr);
-    block_len = gifgetbyte(grr);
-  }
-  return gfex != 0;
+    if (!gfex)
+        Gif_DeleteArray(data);
+    while (block_len > 0) {
+        uint8_t buffer[GIF_MAX_BLOCK];
+        gifgetblock(buffer, block_len, grr);
+        block_len = gifgetbyte(grr);
+    }
+    return gfex != 0;
 }
 
 
@@ -704,7 +738,9 @@ read_application_extension(Gif_Context *gfc, int position, Gif_Reader *grr)
   gifgetblock(buffer, len, grr);
 
   /* Read the Netscape loop extension. */
-  if (len == 11 && memcmp(buffer, "NETSCAPE2.0", 11) == 0) {
+  if (len == 11
+      && (memcmp(buffer, "NETSCAPE2.0", 11) == 0
+          || memcmp(buffer, "ANIMEXTS1.0", 11) == 0)) {
 
     len = gifgetbyte(grr);
     if (len == 3) {
@@ -722,10 +758,8 @@ read_application_extension(Gif_Context *gfc, int position, Gif_Reader *grr)
     }
     return 1;
 
-  } else {
-    buffer[len] = 0;
-    return read_unknown_extension(gfs, 0xFF, (char *)buffer, position, grr);
-  }
+  } else
+    return read_unknown_extension(gfs, 0xFF, (char*)buffer, len, position, grr);
 }
 
 
@@ -747,11 +781,10 @@ read_comment_extension(Gif_Image *gfi, Gif_Reader *grr)
 
 static Gif_Stream *
 read_gif(Gif_Reader *grr, int read_flags,
-	 Gif_ReadErrorHandler handler, void *handler_thunk)
+	 const char* landmark, Gif_ReadErrorHandler handler)
 {
   Gif_Stream *gfs;
   Gif_Image *gfi;
-  Gif_Image *new_gfi;
   Gif_Context gfc;
   int extension_position = 0;
   int unknown_block_type = 0;
@@ -766,13 +799,15 @@ read_gif(Gif_Reader *grr, int read_flags,
 
   gfs = Gif_NewStream();
   gfi = Gif_NewImage();
+  gfs->landmark = landmark;
 
   gfc.stream = gfs;
   gfc.prefix = Gif_NewArray(Gif_Code, GIF_MAX_CODE);
   gfc.suffix = Gif_NewArray(uint8_t, GIF_MAX_CODE);
   gfc.length = Gif_NewArray(uint16_t, GIF_MAX_CODE);
   gfc.handler = handler;
-  gfc.handler_thunk = handler_thunk;
+  gfc.gfi = gfi;
+  gfc.errors[0] = gfc.errors[1] = 0;
 
   if (!gfs || !gfi || !gfc.prefix || !gfc.suffix || !gfc.length)
     goto done;
@@ -793,16 +828,16 @@ read_gif(Gif_Reader *grr, int read_flags,
 
       gfi->identifier = last_name;
       last_name = 0;
-      if (!read_image(grr, &gfc, gfi, read_flags)
-	  || !Gif_AddImage(gfs, gfi)) {
-	Gif_DeleteImage(gfi);
-	goto done;
+      if (!Gif_AddImage(gfs, gfi))
+          goto done;
+      else if (!read_image(grr, &gfc, gfi, read_flags)) {
+          Gif_RemoveImage(gfs, gfs->nimages - 1);
+          goto done;
       }
 
-      new_gfi = Gif_NewImage();
-      if (!new_gfi) goto done;
-
-      gfi = new_gfi;
+      gfc.gfi = gfi = Gif_NewImage();
+      if (!gfi)
+          goto done;
       extension_position++;
       break;
 
@@ -832,7 +867,7 @@ read_gif(Gif_Reader *grr, int read_flags,
 	break;
 
        default:
-	read_unknown_extension(gfs, block, 0, extension_position, grr);
+        read_unknown_extension(gfs, block, 0, 0, extension_position, grr);
 	break;
 
       }
@@ -865,11 +900,16 @@ read_gif(Gif_Reader *grr, int read_flags,
   Gif_DeleteArray(gfc.suffix);
   Gif_DeleteArray(gfc.length);
 
-  if (gfs && gfs->errors == 0 && !(read_flags & GIF_READ_TRAILING_GARBAGE_OK) && !grr->eofer(grr)) {
-    gif_read_error(&gfc, 0, "trailing garbage after GIF ignored");
-    /* but clear error count, since the GIF itself was all right */
-    gfs->errors = 0;
+  if (gfs)
+    gfs->errors = gfc.errors[1];
+  if (gfs && gfc.errors[1] == 0
+      && !(read_flags & GIF_READ_TRAILING_GARBAGE_OK)
+      && !grr->eofer(grr)) {
+      gfc.gfi = 0;
+      gif_read_error(&gfc, 0, "trailing garbage after GIF ignored");
   }
+  /* finally, export last message */
+  gif_read_error(&gfc, -1, 0);
 
   return gfs;
 }
@@ -877,7 +917,7 @@ read_gif(Gif_Reader *grr, int read_flags,
 
 Gif_Stream *
 Gif_FullReadFile(FILE *f, int read_flags,
-		 Gif_ReadErrorHandler h, void *hthunk)
+		 const char* landmark, Gif_ReadErrorHandler h)
 {
   Gif_Reader grr;
   if (!f) return 0;
@@ -887,19 +927,19 @@ Gif_FullReadFile(FILE *f, int read_flags,
   grr.block_getter = file_block_getter;
   grr.offseter = file_offseter;
   grr.eofer = file_eofer;
-  return read_gif(&grr, read_flags, h, hthunk);
+  return read_gif(&grr, read_flags, landmark, h);
 }
 
 Gif_Stream *
 Gif_FullReadRecord(const Gif_Record *gifrec, int read_flags,
-		   Gif_ReadErrorHandler h, void *hthunk)
+		   const char* landmark, Gif_ReadErrorHandler h)
 {
   Gif_Reader grr;
   if (!gifrec) return 0;
   make_data_reader(&grr, gifrec->data, gifrec->length);
   if (read_flags & GIF_READ_CONST_RECORD)
     read_flags |= GIF_READ_COMPRESSED;
-  return read_gif(&grr, read_flags, h, hthunk);
+  return read_gif(&grr, read_flags, landmark, h);
 }
 
 
@@ -918,6 +958,10 @@ Gif_ReadRecord(const Gif_Record *gifrec)
   return Gif_FullReadRecord(gifrec, GIF_READ_UNCOMPRESSED, 0, 0);
 }
 
+void
+Gif_SetErrorHandler(Gif_ReadErrorHandler handler) {
+    default_error_handler = handler;
+}
 
 #ifdef __cplusplus
 }
