@@ -1,6 +1,6 @@
 /* -*- mode: c; c-basic-offset: 2 -*- */
 /* gifwrite.c - Functions to write GIFs.
-   Copyright (C) 1997-2013 Eddie Kohler, kohler@cs.ucla.edu
+   Copyright (C) 1997-2014 Eddie Kohler, ekohler@gmail.com
    This file is part of the LCDF GIF library.
 
    The LCDF GIF library is free software. It is distributed under the GNU
@@ -227,7 +227,11 @@ gfc_define(Gif_CodeTable *gfc, Gif_Node *work_node, uint8_t suffix,
 static inline const uint8_t *
 gif_imageline(Gif_Image *gfi, unsigned pos)
 {
-  unsigned y = pos / gfi->width, x = pos - y * gfi->width;
+  unsigned y, x;
+  if (gfi->width == 0)
+    return NULL;
+  y = pos / gfi->width;
+  x = pos - y * gfi->width;
   if (y == (unsigned) gfi->height)
     return NULL;
   else if (!gfi->interlace)
@@ -509,7 +513,7 @@ save_compression_result(Gif_Image *gfi, Gif_Writer *grr, int ok)
     if (ok) {
       gfi->compressed = grr->v;
       gfi->compressed_len = grr->pos;
-      gfi->free_compressed = Gif_DeleteArrayFunc;
+      gfi->free_compressed = Gif_Free;
       grr->v = 0;
       grr->cap = 0;
     } else
@@ -667,7 +671,7 @@ write_image(Gif_Stream *gfs, Gif_Image *gfi, Gif_CodeTable *gfc,
     }
 
   } else if (!gfi->img) {
-    Gif_UncompressImage(gfi);
+    Gif_UncompressImage(gfs, gfi);
     write_compressed_data(gfi, min_code_bits, gfc, grr);
     Gif_ReleaseUncompressedImage(gfi);
 
@@ -781,21 +785,24 @@ write_generic_extension(Gif_Extension *gfex, Gif_Writer *grr)
   gifputbyte('!', grr);
   gifputbyte(gfex->kind, grr);
   if (gfex->kind == 255) {	/* an application extension */
-    int len = gfex->application ? strlen(gfex->application) : 0;
-    if (len) {
-      gifputbyte(len, grr);
-      gifputblock((const uint8_t *)gfex->application, len, grr);
+    if (gfex->applength) {
+      gifputbyte(gfex->applength, grr);
+      gifputblock((const uint8_t*) gfex->appname, gfex->applength, grr);
     }
   }
-  while (pos + 255 < gfex->length) {
-    gifputbyte(255, grr);
-    gifputblock(gfex->data + pos, 255, grr);
-    pos += 255;
-  }
-  if (pos < gfex->length) {
-    uint32_t len = gfex->length - pos;
-    gifputbyte(len, grr);
-    gifputblock(gfex->data + pos, len, grr);
+  if (gfex->packetized)
+    gifputblock(gfex->data, gfex->length, grr);
+  else {
+    while (pos + 255 < gfex->length) {
+      gifputbyte(255, grr);
+      gifputblock(gfex->data + pos, 255, grr);
+      pos += 255;
+    }
+    if (pos < gfex->length) {
+      uint32_t len = gfex->length - pos;
+      gifputbyte(len, grr);
+      gifputblock(gfex->data + pos, len, grr);
+    }
   }
   gifputbyte(0, grr);
 }
