@@ -1,5 +1,5 @@
 /* support.c - Support functions for gifsicle.
-   Copyright (C) 1997-2013 Eddie Kohler, ekohler@gmail.com
+   Copyright (C) 1997-2014 Eddie Kohler, ekohler@gmail.com
    This file is part of gifsicle.
 
    Gifsicle is free software. It is distributed under the GNU Public License,
@@ -23,106 +23,93 @@ int no_warnings = 0;
 
 
 static void
-verror(int need_file, int seriousness, const char *fmt, va_list val)
+verror(const char* landmark, int need_file,
+       int seriousness, const char *fmt, va_list val)
 {
-  char pattern[BUFSIZ];
-  char buffer[BUFSIZ];
-  static char *printed_file = 0;
-  static int just_printed_context = 0;
-  const char *initial_prefix = program_name;
-  const char *prefix = "";
-  const char *iname = input_name;
-  if (!iname)
-    iname = "<stdin>";
-  if (mode == EXPLODING && active_output_data.active_output_name)
-    iname = active_output_data.active_output_name;
+    char pbuf[256], buf[BUFSIZ], xbuf[BUFSIZ];
+    const char* xfmt;
+    int n, i, p;
+    size_t xi;
 
-  if (printed_file && strcmp(printed_file, iname) != 0) {
-      free(printed_file);
-      printed_file = 0;
-  }
-  if (need_file == 2 && !printed_file)
-      initial_prefix = iname;
-  else if (need_file && !printed_file) {
-      if (mode != BLANK_MODE && mode != MERGING && nested_mode != MERGING) {
-	  fprintf(stderr, "%s: While processing '%s':\n", program_name, iname);
-	  just_printed_context = 1;
-	  prefix = "  ";
-	  initial_prefix = "";
-      }
-      printed_file = malloc(strlen(iname) + 1);
-      strcpy(printed_file, iname);
-  } else if (just_printed_context && seriousness == 0) {
-      prefix = "  ";
-      initial_prefix = "";
-  } else
-      just_printed_context = 0;
+    if (!fmt || !*fmt)
+        return;
 
-  if (seriousness > 2)
-      sprintf(pattern, "%s%s%s fatal error: %%s\n",
-	      initial_prefix, *initial_prefix ? ":" : "", prefix);
-  else if (seriousness == 1)
-      sprintf(pattern, "%s%s%s warning: %%s\n",
-	      initial_prefix, *initial_prefix ? ":" : "", prefix);
-  else
-      sprintf(pattern, "%s%s%s %%s\n",
-	      initial_prefix, *initial_prefix ? ":" : "", prefix);
+    if (!landmark && need_file && active_output_data.active_output_name
+        && mode != BLANK_MODE && mode != MERGING && nested_mode != MERGING)
+        landmark = active_output_data.active_output_name;
+    else if (!landmark)
+        landmark = "";
 
-  if (seriousness > 1)
-    error_count++;
-  else if (no_warnings)
-    return;
+    if (seriousness > 2)
+        xfmt = "%s:%s%s fatal error: ";
+    else if (seriousness == 1)
+        xfmt = "%s:%s%s warning: ";
+    else
+        xfmt = "%s:%s%s ";
+    snprintf(pbuf, sizeof(pbuf), xfmt, program_name, landmark, *landmark ? ":" : "");
+    p = strlen(pbuf);
 
-  /* try and keep error messages together (no interleaving of error messages
-     from two gifsicle processes in the same command line) by calling fprintf
-     only once */
-  verbose_endline();
-  if (strlen(fmt) + strlen(pattern) < BUFSIZ) {
-    sprintf(buffer, pattern, fmt);
-    vfprintf(stderr, buffer, val);
-  } else {
-    pattern[strlen(pattern) - 3] = 0;
-    fprintf(stderr, "%s", pattern);
-    vfprintf(stderr, fmt, val);
-    putc('\n', stderr);
-  }
+    Clp_vsnprintf(clp, buf, sizeof(buf), fmt, val);
+    n = strlen(buf);
+    if ((size_t) n + 1 < sizeof(buf) && (n == 0 || buf[n - 1] != '\n')) {
+        buf[n++] = '\n';
+        buf[n] = 0;
+    }
+
+    xi = 0;
+    for (i = 0; i != n; ) {
+        /* char* pos = (char*) memchr(&buf[i], '\n', n - i);
+           int l = (pos ? &pos[1] - &buf[i] : n - i); */
+        int l = n - i;
+        int xd = snprintf(&xbuf[xi], sizeof(xbuf) - xi, "%.*s%.*s",
+                          (int) p, pbuf, (int) l, &buf[i]);
+        i += l;
+        xi = (xi + xd > sizeof(xbuf) ? sizeof(xbuf) : xi + xd);
+    }
+
+    if (seriousness == 1 && no_warnings)
+        return;
+    else if (seriousness > 1)
+        error_count++;
+
+    verbose_endline();
+    fwrite(xbuf, 1, xi, stderr);
 }
 
-void
-fatal_error(const char *message, ...)
-{
-  va_list val;
-  va_start(val, message);
-  verror(0, 3, message, val);
-  va_end(val);
-  exit(EXIT_USER_ERR);
+void fatal_error(const char* format, ...) {
+    va_list val;
+    va_start(val, format);
+    verror((const char*) 0, 0, 3, format, val);
+    va_end(val);
+    exit(EXIT_USER_ERR);
 }
 
-void
-error(int need_file, const char *message, ...)
-{
-  va_list val;
-  va_start(val, message);
-  verror(need_file, 2, message, val);
-  va_end(val);
+void lerror(const char* landmark, const char* format, ...) {
+    va_list val;
+    va_start(val, format);
+    verror(landmark, 2, 2, format, val);
+    va_end(val);
 }
 
-void
-warning(int need_file, const char *message, ...)
-{
-  va_list val;
-  va_start(val, message);
-  verror(need_file, 1, message, val);
-  va_end(val);
+void error(int need_file, const char* format, ...) {
+    va_list val;
+    va_start(val, format);
+    verror((const char*) 0, need_file, 2, format, val);
+    va_end(val);
 }
 
-void
-warncontext(int need_file, const char *message, ...)
-{
-  va_list val;
-  va_start(val, message);
-  verror(need_file, 0, message, val);
-  va_end(val);
+void lwarning(const char* landmark, const char* format, ...) {
+    va_list val;
+    va_start(val, format);
+    verror(landmark, 2, 1, format, val);
+    va_end(val);
+}
+
+void warning(int need_file, const char* format, ...) {
+    va_list val;
+    va_start(val, format);
+    verror((const char*) 0, need_file, 1, format, val);
+    va_end(val);
 }
 
 void
@@ -229,9 +216,12 @@ Whole-GIF options: Also --no-OPTION.\n\
   printf("\
       --resize WxH              Resize the output GIF to WxH.\n\
       --resize-width W          Resize to width W and proportional height.\n\
-      --resize-height H         Resize to height H and proportional width.\n");
+      --resize-height H         Resize to height H and proportional width.\n\
+      --resize-fit WxH          Resize if necessary to fit within WxH.\n");
   printf("\
       --scale XFACTOR[xYFACTOR] Scale the output GIF by XFACTORxYFACTOR.\n\
+      --resize-method METHOD    Set resizing method.\n\
+      --resize-colors N         Resize can add new colors up to N.\n\
       --transform-colormap CMD  Transform each output colormap by shell CMD.\n\
       --use-colormap CMAP       Set output GIF's colormap to CMAP, which can\n\
                                 be 'web', 'gray', 'bw', or a GIF file.\n\n");
@@ -372,7 +362,7 @@ extension_info(FILE *where, Gif_Stream *gfs, Gif_Extension *gfex, int count)
   fprintf(where, "  extension %d: ", count);
   if (gfex->kind == 255) {
     fprintf(where, "app '");
-    safe_puts(gfex->application, strlen(gfex->application), where);
+    safe_puts(gfex->appname, gfex->applength, where);
     fprintf(where, "'");
   } else {
     if (gfex->kind >= 32 && gfex->kind < 127)
@@ -381,9 +371,12 @@ extension_info(FILE *where, Gif_Stream *gfs, Gif_Extension *gfex, int count)
       fprintf(where, "0x%02X", gfex->kind);
   }
   if (gfex->position >= gfs->nimages)
-    fprintf(where, " at end\n");
+    fprintf(where, " at end");
   else
-    fprintf(where, " before #%d\n", gfex->position);
+    fprintf(where, " before #%d", gfex->position);
+  if (gfex->packetized)
+    fprintf(where, " packetized");
+  fprintf(where, "\n");
 
   /* Now, hexl the data. */
   while (len > 0) {
@@ -637,7 +630,7 @@ parse_dimensions(Clp_Parser *clp, const char *arg, int complain, void *thunk)
   }
 
   if (complain)
-    return Clp_OptionError(clp, "invalid dimensions '%s' (want WxH)", arg);
+    return Clp_OptionError(clp, "invalid dimensions %<%s%> (want WxH)", arg);
   else
     return 0;
 }
@@ -656,20 +649,35 @@ parse_position(Clp_Parser *clp, const char *arg, int complain, void *thunk)
   }
 
   if (complain)
-    return Clp_OptionError(clp, "invalid position '%s' (want 'X,Y')", arg);
+    return Clp_OptionError(clp, "invalid position %<%s%> (want 'X,Y')", arg);
   else
     return 0;
 }
 
+static double strtod_fraction(const char* arg, char** endptr) {
+    char* val, *val2;
+    double d = strtod(arg, &val);
+    if (arg != val && *val == '/') {
+        double denom = strtod(val + 1, &val2);
+        if (val2 != val + 1 && denom) {
+            d /= denom;
+            val = val2;
+        }
+    }
+    if (endptr)
+        *endptr = val;
+    return d;
+}
+
 int
 parse_scale_factor(Clp_Parser *clp, const char *arg, int complain, void *thunk)
 {
   char *val;
   (void)thunk;
 
-  parsed_scale_factor_x = strtod(arg, &val);
+  parsed_scale_factor_x = strtod_fraction(arg, &val);
   if (*val == 'x') {
-    parsed_scale_factor_y = strtod(val + 1, &val);
+    parsed_scale_factor_y = strtod_fraction(val + 1, &val);
     if (*val == 0)
       return 1;
   } else if (*val == 0) {
@@ -678,7 +686,7 @@ parse_scale_factor(Clp_Parser *clp, const char *arg, int complain, void *thunk)
   }
 
   if (complain)
-    return Clp_OptionError(clp, "invalid scale factor '%s' (want XxY)", arg);
+    return Clp_OptionError(clp, "invalid scale factor %<%s%> (want XxY)", arg);
   else
     return 0;
 }
@@ -719,7 +727,7 @@ parse_rectangle(Clp_Parser *clp, const char *arg, int complain, void *thunk)
   }
 
   if (complain)
-    return Clp_OptionError(clp, "invalid rectangle '%s' (want 'X1,Y1-X2,Y2' or 'X1,Y1+WxH'", input_arg);
+    return Clp_OptionError(clp, "invalid rectangle %<%s%> (want X1,Y1-X2,Y2 or X1,Y1+WxH", input_arg);
   else
     return 0;
 }
@@ -767,7 +775,7 @@ parse_color(Clp_Parser *clp, const char *arg, int complain, void *thunk)
     if (len == 0 || len % 3 != 0
 	|| (int)strspn(arg, "0123456789ABCDEFabcdef") != len) {
       if (complain)
-	Clp_OptionError(clp, "invalid color '%s' (want '#RGB' or '#RRGGBB')",
+	Clp_OptionError(clp, "invalid color %<%s%> (want #RGB or #RRGGBB)",
 			input_arg);
       return 0;
     }
@@ -812,7 +820,7 @@ parse_color(Clp_Parser *clp, const char *arg, int complain, void *thunk)
 
  error:
   if (complain)
-    return Clp_OptionError(clp, "invalid color '%s'", input_arg);
+    return Clp_OptionError(clp, "invalid color %<%s%>", input_arg);
   else
     return 0;
 }
@@ -827,7 +835,7 @@ parse_two_colors(Clp_Parser *clp, const char *arg, int complain, void *thunk)
 
   arg = Clp_Shift(clp, 0);
   if (!arg && complain)
-    return Clp_OptionError(clp, "'%O' takes two color arguments");
+    return Clp_OptionError(clp, "%<%O%> takes two color arguments");
   else if (!arg)
     return 0;
 
@@ -883,7 +891,7 @@ read_text_colormap(FILE *f, const char *name)
       if (green > 255) green = 255;
       if (blue > 255) blue = 255;
       if (ncol >= 256) {
-	error(0, "%s: maximum 256 colors allowed in colormap", name);
+	lerror(name, "maximum 256 colors allowed in colormap");
 	break;
       } else {
 	col[ncol].gfc_red = red;
@@ -902,7 +910,7 @@ read_text_colormap(FILE *f, const char *name)
   }
 
   if (ncol == 0) {
-    error(0, "%s: file not in colormap format", name);
+    lerror(name, "file not in colormap format");
     Gif_DeleteColormap(cm);
     return 0;
   } else {
@@ -911,6 +919,12 @@ read_text_colormap(FILE *f, const char *name)
   }
 }
 
+static void
+no_gifread_error(Gif_Stream* gfs, Gif_Image* gfi,
+                 int is_error, const char *message) {
+    (void) gfs, (void) gfi, (void) is_error, (void) message;
+}
+
 Gif_Colormap *
 read_colormap_file(const char *name, FILE *f)
 {
@@ -927,7 +941,7 @@ read_colormap_file(const char *name, FILE *f)
     else
       f = fopen(name, "rb");
     if (!f) {
-      error(0, "%s: %s", name, strerror(errno));
+      lerror(name, "%s", name, strerror(errno));
       return 0;
     }
   }
@@ -938,15 +952,16 @@ read_colormap_file(const char *name, FILE *f)
   c = getc(f);
   ungetc(c, f);
   if (c == 'G') {
-    Gif_Stream *gfs = Gif_ReadFile(f);
+    Gif_Stream *gfs = Gif_FullReadFile(f, GIF_READ_COMPRESSED, 0, no_gifread_error);
     if (!gfs)
-      error(0, "%s: file not in GIF format", name);
-    else if (!gfs->global)
-      error(0, "%s: can't use as palette (no global color table)", name);
+      lerror(name, "file not in GIF format");
+    else if (!gfs->global
+             && (gfs->nimages == 0 || !gfs->images[0]->local))
+      lerror(name, "can%,t use as palette (no global color table)");
     else {
       if (gfs->errors)
-	warning(0, "%s: there were errors reading this GIF", name);
-      cm = Gif_CopyColormap(gfs->global);
+	lwarning(name, "there were errors reading this GIF");
+      cm = Gif_CopyColormap(gfs->global ? gfs->global : gfs->images[0]->local);
     }
 
     Gif_DeleteStream(gfs);
@@ -1000,18 +1015,13 @@ clear_def_frame_once_options(void)
 
 
 Gt_Frame *
-add_frame(Gt_Frameset *fset, int number, Gif_Stream *gfs, Gif_Image *gfi)
+add_frame(Gt_Frameset *fset, Gif_Stream *gfs, Gif_Image *gfi)
 {
-  if (number < 0) {
-    while (fset->count >= fset->cap) {
-      fset->cap *= 2;
-      Gif_ReArray(fset->f, Gt_Frame, fset->cap);
+    int number = fset->count++;
+    while (number >= fset->cap) {
+        fset->cap *= 2;
+        Gif_ReArray(fset->f, Gt_Frame, fset->cap);
     }
-    number = fset->count++;
-  } else {
-    assert(number < fset->count);
-    blank_frameset(fset, number, number, 0);
-  }
 
   /* Mark the stream and the image both */
   gfs->refcount++;
@@ -1029,17 +1039,18 @@ add_frame(Gt_Frameset *fset, int number, Gif_Stream *gfs, Gif_Image *gfi)
 static Gif_Extension *
 copy_extension(Gif_Extension *src)
 {
-  Gif_Extension *dest = Gif_NewExtension(src->kind, src->application);
-  if (!dest) return 0;
-  dest->data = Gif_NewArray(uint8_t, src->length);
-  dest->length = src->length;
-  dest->free_data = Gif_DeleteArrayFunc;
-  if (!dest->data) {
-    Gif_DeleteExtension(dest);
-    return 0;
-  }
-  memcpy(dest->data, src->data, src->length);
-  return dest;
+    Gif_Extension *dest = Gif_NewExtension(src->kind, src->appname, src->applength);
+    if (!dest) return 0;
+    dest->data = Gif_NewArray(uint8_t, src->length);
+    dest->length = src->length;
+    dest->free_data = Gif_Free;
+    if (!dest->data) {
+        Gif_DeleteExtension(dest);
+        return 0;
+    }
+    memcpy(dest->data, src->data, src->length);
+    dest->packetized = src->packetized;
+    return dest;
 }
 
 
@@ -1104,14 +1115,14 @@ find_color_or_error(Gif_Color *color, Gif_Stream *gfs, Gif_Image *gfi,
       return color->pixel;
     else {
       if (color_context)
-	  error(0, "%s color out of range", color_context);
+	  lerror(gfs->landmark, "%s color out of range", color_context);
       return -1;
     }
   }
 
   index = Gif_FindColor(gfcm, color);
   if (index < 0 && color_context)
-    error(2, "%s color not in colormap", color_context);
+    lerror(gfs->landmark, "%s color not in colormap", color_context);
   return index;
 }
 
@@ -1133,12 +1144,13 @@ set_background(Gif_Stream *gfs, Gt_OutputData *output_data)
     if (output_data->background.haspixel) {
 	if (gfs->images[0]->transparent >= 0) {
 	    static int context = 0;
-	    warning(1, "irrelevant background color");
-	    if (!context) {
-		warncontext(1, "(The background will appear transparent because");
-		warncontext(1, "the first image contains transparency.)");
-		context = 1;
-	    }
+            if (!context) {
+                warning(1, "irrelevant background color\n"
+                        "  (The background will appear transparent because"
+                        "  the first image contains transparency.)");
+                context = 1;
+            } else
+                warning(1, "irrelevant background color");
 	}
 	background = output_data->background;
 	goto search;
@@ -1163,7 +1175,8 @@ set_background(Gif_Stream *gfs, Gt_OutputData *output_data)
 	    else if (original_bg_transparent)
 		want_transparent = 1;
 	    else if (merger[i]->transparent.haspixel) {
-		if (background.haspixel && !GIF_COLOREQ(&background, &merger[i]->transparent))
+		if (background.haspixel
+                    && !GIF_COLOREQ(&background, &merger[i]->transparent))
 		    conflict = 1;
 		else {
 		    background = merger[i]->transparent;
@@ -1177,11 +1190,12 @@ set_background(Gif_Stream *gfs, Gt_OutputData *output_data)
     /* Report conflicts. */
     if (conflict || (want_transparent && gfs->images[0]->transparent < 0)) {
 	static int context = 0;
-	warning(1, "input images have conflicting background colors");
-	if (!context) {
-	  warncontext(1, "(This means some animation frames may appear incorrect.)");
-	  context = 1;
-	}
+        if (!context) {
+            warning(1, "input images have conflicting background colors\n"
+                    "  (This means some animation frames may appear incorrect.)");
+            context = 1;
+        } else
+            warning(1, "input images have conflicting background colors");
     }
 
     /* If no important background color, bag. */
@@ -1279,18 +1293,21 @@ analyze_crop(int nmerger, Gt_Crop *crop, int compress_immediately)
 {
   int i, nframes = 0;
   int l = 0x7FFFFFFF, r = 0, t = 0x7FFFFFFF, b = 0;
+  Gif_Stream* cropped_gfs = 0;
 
   /* count frames to which this crop applies */
   for (i = 0; i < nmerger; i++)
-    if (merger[i]->crop == crop)
-      nframes++;
+      if (merger[i]->crop == crop) {
+          cropped_gfs = merger[i]->stream;
+          nframes++;
+      }
 
   /* find border of frames */
   for (i = 0; i < nmerger; i++)
     if (merger[i]->crop == crop) {
       Gt_Frame *fr = merger[i];
       int ll, tt, rr, bb;
-      if (nframes <= 1) {
+      if (!fr->position_is_offset) {
 	ll = fr->image->left;
 	tt = fr->image->top;
 	rr = fr->image->left + fr->image->width;
@@ -1322,10 +1339,11 @@ analyze_crop(int nmerger, Gt_Crop *crop, int compress_immediately)
   crop->top_offset = crop->y;
   if (crop->x < 0 || crop->y < 0 || crop->w <= 0 || crop->h <= 0
       || crop->x + crop->w > r || crop->y + crop->h > b) {
-    error(1, "cropping dimensions don't fit image");
-    crop->ready = 2;
+      lerror(cropped_gfs ? cropped_gfs->landmark : (const char*) 0,
+             "cropping dimensions don%,t fit image");
+      crop->ready = 2;
   } else
-    crop->ready = 1;
+      crop->ready = 1;
 
   /* Remove transparent edges. */
   if (crop->transparent_edges && crop->ready == 1) {
@@ -1347,7 +1365,7 @@ analyze_crop(int nmerger, Gt_Crop *crop, int compress_immediately)
 	if (srci->transparent >= 0) {
 	  int x, y;
 	  uint8_t **img;
-	  Gif_UncompressImage(srci);
+	  Gif_UncompressImage(fr->stream, srci);
 	  img = srci->img;
 
 	  /* Move top edge down over transparency */
@@ -1434,6 +1452,24 @@ apply_frame_transparent(Gif_Image *gfi, Gt_Frame *fr)
     return old_transparent;
 }
 
+static void mark_used_background_color(Gt_Frame* fr) {
+    Gif_Stream* gfs = fr->stream;
+    Gif_Image* gfi = fr->image;
+    if ((fr->transparent.haspixel != 0
+         ? fr->transparent.haspixel != 255
+         : gfi->transparent < 0)
+        && ((fr->disposal >= 0
+             ? fr->disposal
+             : gfi->disposal) == GIF_DISPOSAL_BACKGROUND
+            || gfi->left != 0
+            || gfi->top != 0
+            || gfi->width != gfs->screen_width
+            || gfi->height != gfs->screen_height)
+        && gfs->global
+        && gfs->background < gfs->global->ncol)
+        gfs->global->col[gfs->background].haspixel |= 1;
+}
+
 Gif_Stream *
 merge_frame_interval(Gt_Frameset *fset, int f1, int f2,
 		     Gt_OutputData *output_data, int compress_immediately,
@@ -1445,10 +1481,13 @@ merge_frame_interval(Gt_Frameset *fset, int f1, int f2,
 
   global->ncol = 0;
   dest->global = global;
+  if (output_data->active_output_name)
+      dest->landmark = output_data->active_output_name;
   /* 11/23/98 A new stream's screen size is 0x0; we'll use the max of the
      merged-together streams' screen sizes by default (in merge_stream()) */
 
-  if (f2 < 0) f2 = fset->count - 1;
+  if (f2 < 0)
+      f2 = fset->count - 1;
   nmerger = 0;
   merger_flatten(fset, f1, f2);
   if (nmerger == 0) {
@@ -1458,9 +1497,10 @@ merge_frame_interval(Gt_Frameset *fset, int f1, int f2,
 
   /* decide whether stream is huge */
   {
-    int s;
+    size_t s;
     for (i = s = 0; i < nmerger; i++)
-      s += ((merger[i]->image->width * merger[i]->image->height) / 1024) + 1;
+        s += (((unsigned) merger[i]->image->width
+               * (unsigned) merger[i]->image->height) / 1024) + 1;
     *huge_stream = (s > 200 * 1024); /* 200 MB */
     if (*huge_stream && !compress_immediately) {
       warning(1, "huge GIF, conserving memory (processing may take a while)");
@@ -1510,6 +1550,7 @@ merge_frame_interval(Gt_Frameset *fset, int f1, int f2,
       mark_used_colors(merger[i]->stream, merger[i]->image, merger[i]->crop,
                        compress_immediately);
       merger[i]->image->transparent = old_transp;
+      mark_used_background_color(merger[i]);
   }
 
   /* copy stream-wide information from output_data */
@@ -1548,7 +1589,7 @@ merge_frame_interval(Gt_Frameset *fset, int f1, int f2,
     if (fr->crop) {
       int preserve_total_crop;
       srci = Gif_CopyImage(fr->image);
-      Gif_UncompressImage(srci);
+      Gif_UncompressImage(fr->stream, srci);
 
       /* Zero-delay frames are a special case.  You might think it was OK to
 	 get rid of totally-cropped delay-0 frames, but many browsers treat
@@ -1565,7 +1606,7 @@ merge_frame_interval(Gt_Frameset *fset, int f1, int f2,
       }
     } else {
       srci = fr->image;
-      Gif_UncompressImage(srci);
+      Gif_UncompressImage(fr->stream, srci);
     }
 
     /* It was pretty stupid to remove this code, which I did between 1.2b6 and
@@ -1581,7 +1622,7 @@ merge_frame_interval(Gt_Frameset *fset, int f1, int f2,
 	|| fr->flip_horizontal || fr->flip_vertical || fr->rotation)
       same_compressed_ok = 0;
 
-    desti = merge_image(dest, fr->stream, srci, same_compressed_ok);
+    desti = merge_image(dest, fr->stream, srci, fr, same_compressed_ok);
 
     srci->transparent = old_transp; /* restore real transparent value */
 
@@ -1629,7 +1670,7 @@ merge_frame_interval(Gt_Frameset *fset, int f1, int f2,
       } else if (desti->compressed)
 	Gif_ReleaseCompressedImage(desti);
     } else if (compress_immediately <= 0) {
-      Gif_UncompressImage(desti);
+      Gif_UncompressImage(dest, desti);
       Gif_ReleaseCompressedImage(desti);
     }
 
@@ -1637,7 +1678,8 @@ merge_frame_interval(Gt_Frameset *fset, int f1, int f2,
     /* 6/17/02 store information about image's background */
     if (fr->stream->images[0]->transparent >= 0)
 	fr->transparent.haspixel = 2;
-    else if (fr->stream->global && fr->stream->background < fr->stream->global->ncol) {
+    else if (fr->stream->global
+             && fr->stream->background < fr->stream->global->ncol) {
 	fr->transparent = fr->stream->global->col[fr->stream->background];
 	fr->transparent.haspixel = 1;
     } else
@@ -1665,22 +1707,8 @@ merge_frame_interval(Gt_Frameset *fset, int f1, int f2,
   }
   /** END MERGE LOOP **/
 
-  /* Cropping the whole output? */
+  /* Cropping the whole output? Reset logical screen */
   if (merger[0]->crop && merger[0]->crop == merger[nmerger - 1]->crop) {
-    /* Adjust positions */
-    int l = 0x7FFFFFFF, t = 0x7FFFFFFF;
-    for (i = 0; i < dest->nimages && (l || t); i++) {
-      Gif_Image *gfi = dest->images[i];
-      if (gfi->left < l)
-        l = gfi->left;
-      if (gfi->top < t)
-        t = gfi->top;
-    }
-    for (i = 0; i < dest->nimages; i++) {
-      Gif_Image *gfi = dest->images[i];
-      gfi->left -= l;
-      gfi->top -= t;
-    }
     /* 13.May.2008: Set the logical screen to the cropped dimensions */
     /* 18.May.2008: Unless --crop-transparency is on */
     if (merger[0]->crop->transparent_edges)
@@ -1701,6 +1729,19 @@ merge_frame_interval(Gt_Frameset *fset, int f1, int f2,
   /* Find the background color in the colormap, or add it if we can */
   set_background(dest, output_data);
 
+  /* The global colormap might be empty even given images that use it, if
+     those images are entirely transparent. Since absent global colormaps
+     are surprising, assign a black-and-white colormap */
+  if (dest->global->ncol == 0) {
+      for (i = 0; i != dest->nimages; ++i)
+          if (!dest->images[i]->local) {
+              GIF_SETCOLOR(&dest->global->col[0], 0, 0, 0);
+              GIF_SETCOLOR(&dest->global->col[1], 255, 255, 255);
+              dest->global->ncol = 2;
+              break;
+          }
+  }
+
   return dest;
 }
 
